/*
 * scanner specification for C, adapted from:
 * https://www.lysator.liu.se/c/ANSI-C-grammar-l.html
 */

D			[0-9]
L			[a-zA-Z_]
H			[a-fA-F0-9]
E			[Ee][\+\-]?{D}+
FS			(f|F|l|L)
IS			(u|U|l|L)*

%{
#include <stdio.h>
#include "c.tab.h"

int lineno = 1;

void comment(void);
int check_type(void);
%}


%%


\/\*			%{ comment(); %}

auto			%{ return AUTO; %}
break			%{ return BREAK; %}
case			%{ return CASE; %}
char			%{ return CHAR; %}
const			%{ return CONST; %}
continue		%{ return CONTINUE; %}
default			%{ return DEFAULT; %}
do			%{ return DO; %}
double			%{ return DOUBLE; %}
else			%{ return ELSE; %}
enum			%{ return ENUM; %}
extern			%{ return EXTERN; %}
float			%{ return FLOAT; %}
for			%{ return FOR; %}
goto			%{ return GOTO; %}
if			%{ return IF; %}
int			%{ return INT; %}
long			%{ return LONG; %}
register		%{ return REGISTER; %}
return			%{ return RETURN; %}
short			%{ return SHORT; %}
signed			%{ return SIGNED; %}
sizeof			%{ return SIZEOF; %}
static			%{ return STATIC; %}
struct			%{ return STRUCT; %}
switch			%{ return SWITCH; %}
typedef			%{ return TYPEDEF; %}
union			%{ return UNION; %}
unsigned		%{ return UNSIGNED; %}
void			%{ return VOID; %}
volatile		%{ return VOLATILE; %}
while			%{ return WHILE; %}

{L}({L}|{D})*		%{ return check_type(); %}

0[xX]{H}+{IS}?		%{ return CONSTANT; %}
0{D}+{IS}?		%{ return CONSTANT; %}
{D}+{IS}?		%{ return CONSTANT; %}
L?'(\\.|[^\\'])+'	%{ return CONSTANT; %}

{D}+{E}{FS}?		%{ return CONSTANT; %}
{D}*\.{D}+({E})?{FS}?	%{ return CONSTANT; %}
{D}+\.{D}*({E})?{FS}?	%{ return CONSTANT; %}

L?\"(\\.|[^\\"])*\"	%{ return STRING_LITERAL; %}

\.\.\.			%{ return ELLIPSIS; %}
\>\>\=			%{ return RIGHT_ASSIGN; %}
\<\<\=			%{ return LEFT_ASSIGN; %}
\+\=			%{ return ADD_ASSIGN; %}
\-\=			%{ return SUB_ASSIGN; %}
\*\=			%{ return MUL_ASSIGN; %}
\/\=			%{ return DIV_ASSIGN; %}
\%\=			%{ return MOD_ASSIGN; %}
\&\=			%{ return AND_ASSIGN; %}
\^\=			%{ return XOR_ASSIGN; %}
\|\=			%{ return OR_ASSIGN; %}
\>\>			%{ return RIGHT_OP; %}
\<\<			%{ return LEFT_OP; %}
\+\+			%{ return INC_OP; %}
\-\-			%{ return DEC_OP; %}
\-\>			%{ return PTR_OP; %}
\&\&			%{ return AND_OP; %}
\|\|			%{ return OR_OP; %}
\<\=			%{ return LE_OP; %}
\>\=			%{ return GE_OP; %}
\=\=			%{ return EQ_OP; %}
\!\=			%{ return NE_OP; %}
\;			%{ return SEMIC; %}
(\{|\<\%)		%{ return LCURL; %}
(\}|\%\>)		%{ return RCURL; %}
\,			%{ return COMMA; %}
\:			%{ return COLON; %}
\=			%{ return ASSIGN; %}
\(			%{ return LPAREN; %}
\)			%{ return RPAREN; %}
(\[|\<\:)		%{ return LBRACK; %}
(\]|\:\>)		%{ return RBRACK; %}
\.			%{ return DOT; %}
\&			%{ return AMPER; %}
\!			%{ return XCLAM; %}
\~			%{ return TILDE; %}
\-			%{ return DASH; %}
\+			%{ return PLUS; %}
\*			%{ return STAR; %}
\/			%{ return SLASH; %}
\%			%{ return PERCENT; %}
\<			%{ return LT_OP; %}
\>			%{ return GT_OP; %}
\^			%{ return CARET; %}
\|			%{ return BAR; %}
\?			%{ return QUEST; %}

\n			%{ lineno++; %}
[\ \t\v\f]		%{ %}

.			%{ /* ignore bad characters */ %}


%%


int yywrap(void) {
  return 1;
}


void comment(void) {
  char c, c1;

  while (1) {
    while ((c = input()) != '*' && c != '\0') ;
    if ((c1 = input()) == '/' || c == '\0') {
      break;
    }
    unput(c1);
  }
}


int check_type(void) {
  /*
   * pseudo code --- this is what it should check
   *
   *	if (scanText == type_name)
   *		return TYPE_NAME;
   *
   *	return IDENTIFIER;
   */

  /*
   *	it actually will only return IDENTIFIER
   */

	return IDENTIFIER;
}


int main(int argc, char *argv[]) {
  int numTokens;
  int token;

  numTokens = 0;
  do {
    token = yylex();
    numTokens++;
  } while (token != 0);
  printf("%d tokens read\n", numTokens);
  return 0;
}
